#ifndef _OPTIONS_H_
#define _OPTIONS_H_
#include <map>
#include <cstdio>

#include <manucgi.h>
using namespace manucgi;

struct OptError: public std::runtime_error {
  OptError(std::string s): std::runtime_error("option error: "+s) {};
};

class Options;
class OptBaseType;
class OptValue;
class OptBaseType {
private:
 public:
  OptBaseType(){};
  virtual ~OptBaseType(){};
  virtual bool parse(Options &opt, 
		     int argc,char *argv[],int &count,OptValue &val)=0;
  virtual bool parse(Options &,const CgiItem &item, OptValue &val)=0;
  virtual std::string usage(){return "";};
  virtual const std::string &name() const {
    static const std::string empty="";
    return empty;};
  virtual std::string url() const {return Cgi::escape_url(name());};
};

class OptValue {
private:
  int mystage;
 public:
  OptValue(int stage=0):mystage(stage){}; 
  virtual ~OptValue(){};
  virtual bool parse(std::string s) {return s=="";};
  virtual void exec(Options &opt){};
  virtual int execution_stage(){return mystage;};
  virtual bool need_value() const {return false;};
  virtual std::string value() const {return "";};
  virtual std::string url(const OptBaseType &base) const;
  virtual std::string form(const OptBaseType &base) const;
};

class OptBool: public OptValue {
  bool &myref;
 public:
  OptBool(bool &ref): myref(ref) {};
  virtual ~OptBool(){};
  virtual bool parse(std::string s) {
    if (s=="" || s=="y" || s=="on" || s=="yes") {myref=true;return true;}
    if (s=="no" || s=="n" || s=="off") {myref=false;return true;}
    return false;
  };
  virtual bool need_value() const {return false;};
  virtual std::string value() const {
    if (myref) return "yes";else return "no";};
};

class OptCases: public OptValue {
  std::map<std::string,int> mymap;
  int &myref;
 public:
  OptCases(char *list[],int *vals,int &ref);
  virtual ~OptCases(){};
  virtual bool parse(std::string s);
  virtual bool need_value() const {return true;};
  virtual std::string value() const;
};

class OptStringRef: public OptValue {
private:
  std::string &myref;
public:
  OptStringRef(std::string &ref):myref(ref){};
  virtual bool parse(std::string s){myref=s;return true;};
  virtual bool need_value() const {return true;};
  virtual std::string value() const {return myref;};
};

class OptDoubleRef: public OptValue {
private:
  double &myref;
public:
  OptDoubleRef(double &ref):myref(ref){};
  virtual bool parse(std::string s){
    if (sscanf(s.c_str(),"%lf",&myref)==1)
      return true;
    return false;
  };
  virtual bool need_value() const {return true;};
  virtual std::string value() const {
    char s[80];
    sprintf(s,"%.2g",myref);
    return std::string(s);
  };
};

class OptIntRef: public OptValue {
private:
  int &myref;
public:
  OptIntRef(int &ref):myref(ref){};
  virtual bool parse(std::string s){
    if (sscanf(s.c_str(),"%d",&myref)==1)
      return true;
    return false;
  };
  virtual bool need_value() const {return true;};
  virtual std::string value() const {
    char s[80];
    sprintf(s,"%d",myref);
    return std::string(s);
  };
};

class OptFun: public OptValue {
  typedef void (*fn_type)(Options &opt);
  fn_type myfn;
  bool active;
public:
  virtual bool parse(std::string s){active=true;return true;};
  OptFun(fn_type fn,int stage=0): OptValue(stage),myfn(fn) {active=false;};
  virtual void exec(Options &opt) {if (active) (*myfn)(opt);};
};

class OptFunForeachString: public OptValue {
private:
  typedef void (*fn_type)(Options &opt,std::string s);
  fn_type myfn;
  std::vector<std::string> list;
public:
  OptFunForeachString(fn_type fn): myfn(fn) {};
  virtual bool parse(std::string s){list.push_back(s); return true;};
  virtual bool need_value() const {return true;};
  virtual void exec(Options &opt);
  virtual std::string url(const OptBaseType &) const;
  virtual std::string form(const OptBaseType &) const;
};

class OptNormal: public OptBaseType {
private:
  std::string myname;
  std::string myalias;
  std::string mydescription;
public:
  OptNormal(std::string name,std::string alias, std::string description):
    myname(name), myalias(alias), mydescription(description) {};
  virtual bool parse(Options &,int argc,char *argv[],int &count,OptValue &val);
  virtual bool parse(Options &,const CgiItem &item, OptValue &val);
  virtual std::string usage();
  virtual const std::string &name() const {return myname;};
};

class OptDefault: public OptNormal {
public:
  OptDefault(std::string name,std::string alias, std::string description);
  virtual bool parse(Options &,int argc,char *argv[],int &count,OptValue &val);
};

class Option {
private:
  OptBaseType *mytype;
  OptValue *myvalue;
public:
  Option(OptBaseType* type, OptValue *value):
    mytype(type), myvalue(value) {};
  ~Option() {
    //andrebbero cancellati, ma quando faccio le copie succedono guai
    //delete mytype;delete myvalue;
  };
  bool parse(Options &op, int argc, char *argv[], int &count);
  bool parse(Options &op, const CgiItem &item);
  int execution_stage() {return myvalue->execution_stage();};
  void exec(Options &opt) {return myvalue->exec(opt);};
  std::string usage() {return mytype->usage();};
  const std::string &name() {return mytype->name();};
  std::string url() const {
    return myvalue->url(*mytype);};
  std::string form() const {
    return myvalue->form(*mytype);};
};

class Options {
  typedef std::vector<Option> list_type;
  list_type list;
 public:
  void add(Option o) {list.push_back(o);};
  void parse(int argc,char *argv[]);
  void parse(Cgi &cgi);
  void exec();
  void usage(std::ostream &out);
  Option &operator[](const std::string &name);
};

#endif
